"""
Alembic migration utilities
"""
import alembic.command
import alembic.config
import contextlib
import logging
import sqlalchemy
import sqlalchemy.exc
import sqlalchemy_utils
import time

#
# Default timeout to 5 seconds
#
LOCK_TIMEOUT = 5

logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter(' * JHHAlchemy - %(message)s'))
logger.addHandler(handler)
logger.setLevel(logging.INFO)

@contextlib.contextmanager
def get_upgrade_lock(dbname, connect_str, timeout=LOCK_TIMEOUT):
    """
    Wait until you can get the lock, then yield it, and eventually release it.

    Inspired by: http://arr.gr/blog/2016/05/mysql-named-locks-in-python-context-managers/

    :param dbname: database to upgrade
    :param connect_str: connection string to the database
    :param timeout: how long to wait between tries for the lock, default 5 seconds
    """
    #
    # Open connection and try to get the lock
    #
    engine = sqlalchemy.create_engine(connect_str)
    cursor = engine.execute("SELECT GET_LOCK('upgrade_{}', {})".format(dbname, timeout))
    lock = cursor.scalar()
    cursor.close()

    #
    # Keep trying until you get it.
    #
    while not lock:
        logger.info('Cannot acquire {} upgrade lock. Sleeping {} seconds.'.format(dbname, timeout))
        time.sleep(timeout)
        cursor = engine.execute("SELECT GET_LOCK('upgrade_{}', {})".format(dbname, timeout))
        lock = cursor.scalar()
        cursor.close()
    logger.info('Acquired {} upgrade lock'.format(dbname))
    yield lock

    #
    # Release the lock and close the connection.
    #
    cursor = engine.execute("SELECT RELEASE_LOCK('upgrade_{}')".format(dbname))
    cursor.close()
    engine.dispose()
    logger.info('Released {} upgrade lock'.format(dbname))


def upgrade(dbname, connect_str, alembic_conf):
    """
    Get the database's upgrade lock and run alembic.

    :param dbname: Name of the database to upgrade/create
    :param connect_str: Connection string to the database (usually Flask's SQLALCHEMY_DATABASE_URI)
    :param alembic_conf: location of alembic.ini
    """
    #
    # The db has to exist before we can get the lock. On the off-chance that another process creates the db between
    # checking if it exists and running the create, ignore the exception.
    #
    if not sqlalchemy_utils.database_exists(connect_str):
        logger.info('Creating {}'.format(dbname))
        try:
            sqlalchemy_utils.create_database(connect_str)
        except sqlalchemy.exc.ProgrammingError as exc:
            if not sqlalchemy_utils.database_exists(connect_str):
                logger.error('Could not create {}'.format(dbname))
                raise exc

    with get_upgrade_lock(dbname, connect_str):
        alembic_config = alembic.config.Config(
            alembic_conf,
            attributes={'configure_logger': False})
        logger.info('Upgrading {} to head'.format(dbname))
        alembic.command.upgrade(alembic_config, 'head')



"""
flask_sqlachemy model mixin for TimeOrder tables
"""
import sqlalchemy


class TimeOrderMixin(object):
    """
    Mixin for tables with time_order columns used for reverse order sorting

    Must be mixed with a model that inherits from jhhalchemy.model
    """
    time_order = sqlalchemy.Column(sqlalchemy.Integer, nullable=False)

    @property
    def timestamp(self):
        """
        Derive the timestamp from time_order.

        :return: the timestamp when the timezone was set
        """
        return -self.time_order

    @timestamp.setter
    def timestamp(self, timestamp):
        """
        Use a timestamp to set time_order.

        :param timestamp: unix timestamp
        """
        self.time_order = -timestamp

    @classmethod
    def read_time_range(cls, *args, **kwargs):
        """
        Get all timezones set within a given time. Uses time_dsc_index

        SELECT *
        FROM <table>
        WHERE time_order <= -<start_timestamp>
        AND time_order >= -<end_timestamp>

        :param args: SQLAlchemy filter criteria, (e.g., uid == uid, type == 1)
        :param kwargs: start_timestamp and end_timestamp are the only kwargs, they specify the range (inclusive)
        :return: model generator
        """
        criteria = list(args)
        start = kwargs.get('start_timestamp')
        end = kwargs.get('end_timestamp')
        if start is not None:
            criteria.append(cls.time_order <= -start)
        if end is not None:
            criteria.append(cls.time_order >= -end)
        return cls.read(*criteria)


"""
Helper Function for APIs that interact with time_order models
"""


class InvalidTimestampRange(Exception):
    """
    Raise this when start_timestamp > end_timestamp
    """
    pass


def get_by_range(model_cls, *args, **kwargs):
    """
    Get ordered list of models for the specified time range.
    The timestamp on the earliest model will likely occur before start_timestamp. This is to ensure that we return
    the models for the entire range.

    :param model_cls: the class of the model to return
    :param args: arguments specific to the model class
    :param kwargs: start_timestamp and end_timestamp (see below) as well as keyword args specific to the model class
    :keyword start_timestamp: the most recent models set before this and all after, defaults to 0
    :keyword end_timestamp: only models set before (and including) this timestamp, defaults to now
    :return: model generator
    """
    start_timestamp = kwargs.get('start_timestamp')
    end_timestamp = kwargs.get('end_timestamp')
    if (start_timestamp is not None) and (end_timestamp is not None) and (start_timestamp > end_timestamp):
        raise InvalidTimestampRange

    models = model_cls.read_time_range(*args, end_timestamp=end_timestamp).order_by(model_cls.time_order)

    #
    # start time -> Loop through until you find one set before or on start
    #
    if start_timestamp is not None:
        index = 0
        for index, model in enumerate(models, start=1):
            if model.timestamp <= start_timestamp:
                break
        models = models[:index]

    return models
